from PySide6.QtCore import Signal, QObject, QUrl, QDir, QByteArray
from PySide6.QtGui import QPixmap, QPixmapCache
from PySide6.QtNetwork import QNetworkAccessManager, QNetworkRequest, QNetworkReply

from .settings import Settings
from .app_config import AppConfig
from .lexemes.lexemes import Lexemes
from .helpers.file_helper import size_f, save_file

from urllib.parse import urlparse

from qasync import asyncSlot
import asyncio

import os
import re
import hashlib
import logging


class ImageDownloader(QObject):  # QObject to allow signal emitting

    # Signal to emit upon single resource was downloaded
    downloaded = Signal(dict)

    # Signal to emit upon all downloading tasks in queue have been finished
    finished = Signal(int)

    RESOURCE_DIR = 'images'

    def __init__(self, base_folder: str = None):

        super().__init__()

        self.settings = Settings(parent=self)

        self.logger = logging.getLogger('image_downloader')

        self.logging = AppConfig.get_logging()
        self.debug = AppConfig.get_debug()

        self.lexemes = Lexemes()

        # Async event loop
        self.event_loop = asyncio.get_event_loop()
        # async download tasks queue
        self.resource_tasks = []

        # Downloaded counter till finished emitted
        self.downloaded_cnt = 0

        # Validate folder and make it QDir
        self.folder = self.get_resource_folder(base_folder)

        if self.debug:
            self.logger.debug(f'Folder to save downloaded resources: {self.folder.path()}')

        os.makedirs(self.folder.path(), exist_ok=True)

        self.network_manager = QNetworkAccessManager()

    def get_resource_folder(self, base_folder: str = None) -> QDir:
        # Check folder and make it QDir
        if base_folder:
            folder = os.path.join(base_folder, self.RESOURCE_DIR)
        else:
            folder = os.path.join(QDir.currentPath(), self.RESOURCE_DIR)
        if isinstance(folder, str):
            folder = QDir(folder)
        return folder

    def download_image(self, url: str) -> None:
        if not self.is_external_url(url):
            if self.debug:
                self.logger.debug(f"Skip downloading as the provided url is not an external url: {url}")
            return

        if self.debug:
            self.logger.debug(f"Downloading {url}")

        request = QNetworkRequest(QUrl(url))
        reply = self.network_manager.get(request)

        # Connecting finished signal
        reply.finished.connect(lambda _reply=reply: self.handle_network_reply(_reply))
        # Connecting error signal
        reply.errorOccurred.connect(lambda error, _reply=reply: self.handle_network_reply(_reply, error))

    def save_image(self, url: str, data: QByteArray) -> None:
        # file_name = os.path.basename(url)
        file_name = self.url_to_filename(url)
        # File path to save
        file_path = os.path.join(self.folder.path(), file_name)
        # Save received data
        if save_file(file_path, data, as_bytearray=True):
            if self.debug:
                self.logger.debug(f"Resource saved {url} to {file_path} [{size_f(len(data))}]")
        else:
            if self.debug:
                self.logger.debug(f"Resource not saved {url} to {file_path} [{size_f(len(data))}]")

    def handle_network_reply(self, reply, error_code=None):
        # Get received status code, say 200
        status_code = reply.attribute(QNetworkRequest.Attribute.HttpStatusCodeAttribute)

        if self.debug:
            self.logger.debug(f'Update response, reply: {reply}, error_code: {error_code}, status_code: {status_code}')

        reply.deleteLater()  # Clean up the QNetworkReply object

        # Make sure there is no error
        if reply.error() == QNetworkReply.NetworkError.NoError:
            # No error, processing the image
            result_message = self.lexemes.get('network_connection_error_empty')
            if self.debug:
                self.logger.debug(result_message)

            # Check if the resource is an image
            mime_type = reply.header(QNetworkRequest.KnownHeaders.ContentTypeHeader)
            if 'image' in mime_type:
                url = reply.url().toString()
                data = reply.readAll()  # type: QByteArray
                if self.debug:
                    self.logger.debug(f"Resource data downloaded {url} [{size_f(len(data))}]")

                # Store the image within the app's cache first
                pixmap = QPixmap()
                pixmap.loadFromData(data)
                QPixmapCache.insert(url, pixmap)

                self.downloaded.emit({'resource_name': url})
                self.downloaded_cnt += 1

                if self.settings.viewer_save_resources:
                    # Saving downloaded files
                    self.save_image(url, data)

                """
                base_name = self.url_to_filename(url)
                file_extension = self.mime_to_extension(mime_type)
                file_name = f"{base_name}{file_extension}"
                """

        elif reply.error() == QNetworkReply.NetworkError.HostNotFoundError:
            # The host was not found, indicating possible DNS issues or no internet connection
            result_message = self.lexemes.get('network_connection_error_connection_or_dns')
        elif reply.error() == QNetworkReply.NetworkError.ConnectionRefusedError:
            # Connection was refused, indicating the server might be down or there are network issues
            result_message = self.lexemes.get('network_connection_error_connection_refused')
        elif reply.error() == QNetworkReply.NetworkError.TimeoutError:
            # The connection timed out, indicating network issues
            result_message = self.lexemes.get('network_connection_error_connection_timed_out')
        elif reply.error() == QNetworkReply.NetworkError.ContentNotFoundError:
            # The requested page or resource is not found
            result_message = self.lexemes.get('network_connection_error_connection_404_error')
        else:
            # Handle other errors
            result_message = self.lexemes.get('network_connection_error_generic_with_status_code',
                                              status_code=status_code)

        if error_code is not None:
            if self.logging:
                self.logger.warning(result_message)
                self.logger.warning(f"Failed to download resource: {reply.errorString()}")

    @asyncSlot()
    async def download_resource_in_queue(self, resource_url) -> None:
        if self.debug:
            self.logger.debug('Downloading resource %d tasks in queue' % len(self.resource_tasks))

        task = asyncio.ensure_future(self.resource_download_async(resource_url))
        # Add task to the local pool first
        self.resource_tasks.append(task)
        # Callback method to set up further actions
        task.add_done_callback(lambda _task: self.resource_task_callback(_task))

        done, pending = await asyncio.wait(
            self.resource_tasks,
            return_when=asyncio.ALL_COMPLETED,  # There is no pending tasks check
            # timeout = 1.5  # to return after the timeout, some task could be pending
        )

        if self.debug:
            self.logger.debug(f'Downloading resource tasks progress. Done {len(done)}, pending {len(pending)}')

        if len(self.resource_tasks) <= 1:
            # Hold this method a particular amount of time to allow completed tasks to settle as this step is
            # typically ahead.
            await asyncio.sleep(0.25, self.event_loop)
            if self.downloaded_cnt > 0:
                # All tasks in queue have finished
                self.finished.emit(self.downloaded_cnt)
                self.downloaded_cnt = 0

    def resource_task_callback(self, task):
        if self.debug:
            self.logger.debug('%s from total %d completed with callback'
                              % (task.get_name(), len(self.resource_tasks)))

        # Remove finished task from the queue
        self.resource_tasks.remove(task)

    async def resource_download_async(self, image_url) -> None:
        """
        Downloading resource which will be processed by resource_downloaded_handler() then.
        """
        self.download_image(image_url)

    async def cancel_tasks(self):
        tasks_total = len(self.resource_tasks)
        for i, task in enumerate(self.resource_tasks):
            if not task.done():
                task_res = task.cancel()
                if self.logging:
                    self.logger.info(
                        f'[{i + 1}/{tasks_total}] Pending task "{task.get_name()}" canceled with result "{task_res}"')
                # Gather all tasks to ensure they are completed before closing
                await asyncio.gather(*self.resource_tasks, return_exceptions=True)

    @staticmethod
    def is_external_url(url, base_domain='example.com'):
        """
        Determine if the given URL is an external URL.
        """
        parsed_url = urlparse(url)
        return not (parsed_url.netloc == '' or parsed_url.netloc.endswith(base_domain))

    @staticmethod
    def url_to_filename(url):
        """
        Generate a filename from a URL by hashing the URL except its basename and combining it with the basename.

        # Convert a URL into a safe filename by replacing non-alphanumeric characters with underscores.
        return re.sub(r'\W+', '_', url)
        """
        # Extract the basename
        basename = os.path.basename(url)

        # Hash the URL excluding the basename
        url_hash = hashlib.sha256(url.replace(basename, '').encode('utf-8')).hexdigest()
        short_hash = url_hash[:8]  # Use only first 8 characters of the hash

        # Combine the short hash with the basename
        safe_basename = basename.replace(os.path.splitext(basename)[1], '')
        safe_basename = re.sub(r'\W+', '_', safe_basename)  # sanitize to make sure it's a valid file name part
        return f"{short_hash}_{safe_basename}{os.path.splitext(basename)[1]}"

    @staticmethod
    def mime_to_extension(mime_type):
        # Map of common image MIME types to file extensions
        mime_map = {
            'image/jpeg': '.jpg',
            'image/png': '.png',
            'image/gif': '.gif',
            'image/bmp': '.bmp',
            'image/svg+xml': '.svg',
            'image/tiff': '.tiff',
            'image/webp': '.webp'
        }
        return mime_map.get(mime_type, '.dat')  # Default to .dat if MIME type is unknown



# Italian lexemes settings_dialog.py
lexemes = {
    # Impostazioni
    "window_title": "Impostazioni",

    "button_close": "Chiudi",

    "tab_general": "Generale",
    "tab_editor_config": "Editore",
    "tab_viewer_config": "Visualizzatore",
    "tab_ai_config": "Configurazione IA",

    "general_app_config_label": "Configurazione dell'app",
    "general_app_language_label": "Lingua",
    "general_app_language_combo_placeholder_text": "Scegli una lingua",
    "general_app_language_combo_accessible_description": "Lingua dell'interfaccia dell'app",
    "general_app_theme_label": "Tema",
    "general_app_theme_combo_placeholder_text": "Scegli un tema",
    "general_app_theme_combo_accessible_description": "Tema dell'interfaccia dell'app",
    "general_app_main_menu_label": "Menu Principale",
    "general_app_main_menu_checkbox": "Mostra il menu principale",
    "general_app_main_menu_checkbox_accessible_description": "Visualizza il menu a discesa principale dell'app",
    "general_app_font_size_label": "Dimensione del carattere: {size}pt",
    "general_app_font_size_slider_accessible_description": "Regola la dimensione del carattere globale dell'app",

    "general_statusbar_label": "Barra di Stato",
    "general_statusbar_show_global_cursor_position_checkbox": "Mostra la Posizione Globale del Cursore",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Visualizza la posizione globale del cursore nella barra di stato",

    "editor_config_label": "Configurazione Editor",
    "editor_config_show_line_numbers_checkbox": "Mostra Numeri di Riga",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Visualizza i numeri di riga nell'editor",

    "viewer_config_label": "Configurazione Visualizzatore",
    "viewer_config_process_emojis_checkbox": "Converti Emoji di Testo in Grafica",
    "viewer_config_process_emojis_checkbox_accessible_description":
        "Converti le emoji di testo in rappresentazioni grafiche",
    "viewer_config_highlight_todos_checkbox": "Evidenzia TODO",
    "viewer_config_highlight_todos_checkbox_accessible_description": "Evidenzia le etichette TODO nel testo",
    "viewer_config_open_link_confirmation_checkbox": "Richiedi Conferma per Aprire i Link",
    "viewer_config_open_link_confirmation_checkbox_accessible_description":
        "Richiedi conferma prima di aprire i link",
    "viewer_config_save_resources_checkbox": "Salva automaticamente le immagini esterne sul disco",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Salva automaticamente copie delle immagini esterne sul disco per l'accesso offline.",

    "ai_config_openai_api_label": "API OpenAI",
    "ai_config_openai_api_url_input_placeholder_text": "URL API",
    "ai_config_openai_api_url_input_accessible_description": "URL dell'API OpenAI",
    "ai_config_openai_api_key_input_placeholder_text": "Chiave API",
    "ai_config_openai_api_key_input_accessible_description": "Chiave API di OpenAI",
    "ai_config_openai_api_supported_models_label": "Modelli supportati",
    "ai_config_ai_model_names_combo_placeholder_text": "Scegli modello",
    "ai_config_ai_model_names_combo_accessible_description": "Modelli supportati da scegliere",

    "ai_config_base_label": "Parametri Base",
    "ai_config_base_system_prompt_label": "Prompt del Sistema",
    "ai_config_base_system_prompt_edit_placeholder_text": "Prompt di sistema base che precede ogni richiesta",
    "ai_config_base_system_prompt_edit_accessible_description":
        "Prompt di sistema base che precede ogni richiesta. Testo semplice.",
    "ai_config_base_response_max_tokens_label": "Massimo di Token di Risposta",
    "ai_config_base_response_max_tokens_input_accessible_description": "Numero massimo di token da ricevere in risposta",
}


from xml.etree import ElementTree

from markdown import Markdown
from markdown.extensions import Extension
from markdown.treeprocessors import Treeprocessor

from .app_config import AppConfig

import re
import logging


class ElementTreeExtension(Extension):
    """
    Custom markdown extension.
    """

    logger = logging.getLogger('tree_extension')

    logging = AppConfig.get_logging()
    debug = AppConfig.get_debug()

    def extendMarkdown(self, md: Markdown) -> None:
        if self.debug:
            self.logger.info('%s extension engaged' % self.__class__.__qualname__)
        # Registering extension
        md.registerExtension(self)
        # noinspection SpellCheckingInspection
        md.treeprocessors.register(ElementTreeProcessor(md), 'notolog_extension', 1)


class ElementTreeProcessor(Treeprocessor):
    """
    Custom markdown extension tree processor.
    """

    logger = logging.getLogger('tree_processor')

    logging = AppConfig.get_logging()
    debug = AppConfig.get_debug()

    def run(self, root):
        # Customize elements tree here
        for idx, elem in enumerate(root.iter()):
            if self.debug:
                self.logger.debug('Process elements > %d: %s' % (idx, elem.tag))
            # Process <table> elements
            if elem.tag == 'table':
                self.process_table(elem)
            if elem.tag in ['th', 'td']:
                self.process_table_td(elem)
            # Process <img> elements
            if elem.tag == 'img':
                self.process_image(elem)
            # Process <p> elements
            if elem.tag == 'p':
                self.process_paragraph(elem)
            # noinspection SpellCheckingInspection
            """
            Works when `codehilite` is switched off:
            if elem.tag == 'pre':
                for id_xy, c_elem in enumerate(elem):
                    if c_elem.tag == 'code':
                        c_elem.set('style', 'background-color: #373737; color: #f5f5f5; padding: 10px;')
            """

    def get_index(self, root, element):
        for idx, child in enumerate(root):
            if element == child:
                return idx
        else:
            raise ValueError("No inner '%s' tag found in '%s'" % (element.tag, root.tag))

    def process_table(self, elem):
        if not elem.get('class'):
            if self.debug:
                self.logger.debug(f'Table without class found {elem}')
            elem.set('class', '_nl_tbl')
            elem.set('cellpadding', '0')
            elem.set('cellspacing', '0')

    def process_table_td(self, elem):
        # Replace 'style' with 'class' as css doesn't work otherwise
        if elem.get('style') and elem.get('style').startswith('text-align'):
            """
            re = QRegularExpression(r'^text-align:\s?(.*?);$')
            match = re.match(elem.get('style'))
            if match.capturedTexts() and match.captured(1):
                ...
            """
            pattern = re.compile(r'^text-align:\s?(.*?);$')
            match = pattern.search(elem.get('style'))
            if match:
                del elem.attrib['style']
                elem.set('class', match.group(1))

    def process_image(self, elem):
        if self.debug:
            self.logger.debug(f'Image element {elem}')
        # elem.set('title', '...')
        # elem.set('onerror', "this.onerror=null;this.src='...';")

    def process_paragraph(self, elem):
        # Check inner elements
        if not (len(elem) > 0):
            return
        # Process inner elements
        for id_xy, c_elem in enumerate(elem):
            if self.debug:
                self.logger.debug('Inner element > %d: %s' % (id_xy, c_elem.tag))
            # Example of how to update elements tree
            if c_elem.tag == 'a' and False:
                # Skip element that was already re-inserted
                if c_elem.get('class') == '_re-inserted':
                    continue
                # Add break space before the element
                br_str = "<br/>"
                br_elem = ElementTree.fromstring(br_str)
                # Remove `a` element
                elem.remove(c_elem)
                # Insert `br` element
                elem.insert(id_xy, br_elem)
                # Setup class as a flag to determine re-inserted `a` element
                c_elem.set('class', '_re-inserted')
                # Re-insert the `a` element with a new class
                elem.insert(id_xy + 1, c_elem)


# tests/test_themes.py

from notolog.helpers.theme_helper import ThemeHelper
from notolog.enums.themes import Themes
from notolog.theme import Theme
from notolog.settings import Settings
from notolog.app_config import AppConfig

from logging import Logger

import os
import pytest


class TestThemeHelper:

    @pytest.fixture(scope="function", autouse=True)
    def test_settings_fixture(self):

        settings = Settings()

        yield settings

    @pytest.fixture(scope="function", autouse=True)
    def test_obj_theme_helper(self, test_settings_fixture, mocker, request):
        """
        Testing object fixture.
        May contain minimal logic for testing purposes.
        """

        # Get the parameter value(s) from the request
        theme = request.param if hasattr(request, 'param') else None

        mocker.patch.object(AppConfig, 'get_debug', return_value=False)
        # Or: monkeypatch.setattr(Settings, 'theme', theme)
        # mocker.patch('app.settings.Settings.app_theme', theme)
        test_settings_fixture.app_theme = theme

        """
        If themes dir is set read themes from there.
        For test purposes to check theme files are readable.
        """
        test_themes_dir = os.path.join(os.path.dirname(__file__), 'test_themes')
        mocker.patch.object(Theme, 'get_themes_dir', return_value=test_themes_dir)

        # Allow to change theme name during test cycle with no caching
        AppConfig.set_test_mode(True)

        yield ThemeHelper()

    @pytest.fixture(scope="function")
    def test_exp_params_fixture(self, request):
        """
        Fixture to pass params and get expected results.
        This method is used only for passing params via pytest fixture.
        """

        # Get the parameter value(s) from the request
        param_values = request.param

        yield param_values

    @pytest.mark.parametrize(
        "test_obj_theme_helper, test_exp_params_fixture",
        [
            ((None), ('default')),
            ((''), ('default')),
            (('default'), ('default')),  # Because of method default()
            (('DEFAULT'), ('default')),
            (('anyval'), ('default')),
        ],
        indirect=True
    )
    def test_theme_helper_init(self, test_obj_theme_helper, test_exp_params_fixture):
        """
        Test and check initial params are exist in the testing object.
        """
        assert isinstance(test_obj_theme_helper.logger, Logger)

        exp_theme = test_exp_params_fixture

        assert hasattr(test_obj_theme_helper, 'theme')
        assert isinstance(test_obj_theme_helper.theme, Theme)
        assert isinstance(test_obj_theme_helper.settings, Settings)
        assert test_obj_theme_helper.theme.theme == exp_theme

    @pytest.mark.parametrize(
        "test_obj_theme_helper, test_exp_params_fixture",
        [
            (None, (None, None, None)),
            ('default', ('test_red', None, 0xFF0000)),
            ('default', ('test_green', True, '#00FF00')),
            ('default', ('test_blue', False, 0x0000FF)),
            ('default', ('test_undefined', None, None)),
            # 'noir_dark' is real as it can be checked with Themes enum
            ('noir_dark', ('test_non_red', False, 0x00FFFF)),
            ('noir_dark', ('test_light', True, '#FFFFFF')),
            ('test_undefined', (None, None, None)),
            ('test_undefined', ('undefined_name', True, None)),
        ],
        indirect=True
    )
    def test_themes_get_color(self, test_obj_theme_helper, test_exp_params_fixture):
        """
        Test getter for correctly loaded and really existent theme's colors.
        """
        param_name, param_css_format, exp_result = test_exp_params_fixture

        assert test_obj_theme_helper.get_color(param_name, param_css_format) == exp_result

    @pytest.mark.parametrize(
        "test_obj_theme_helper, test_exp_params_fixture",
        [
            (None, (None, None)),
            ('default', ('test_smth', None)),
            ('default', ('test_styles', 'QTextEdit {}')),
            ('noir_dark', ('test_smth', 'QTextEdit {}')),
        ],
        indirect=True
    )
    def test_themes_get_css(self, test_obj_theme_helper, test_exp_params_fixture):
        """
        Test getter for correctly loaded and really existent theme's css files.
        """
        param_css_name, exp_result = test_exp_params_fixture

        assert test_obj_theme_helper.get_css(param_css_name) == exp_result


from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QTextCursor, QTextBlockUserData, QTextBlock, QFont
from PySide6.QtWidgets import QPlainTextEdit

from typing import Union

from .app_config import AppConfig
from .text_block_data import TextBlockData

import logging


class EditWidget(QPlainTextEdit):
    # Class to extend or to modify of the original QPlainTextEdit

    content_set = Signal()

    def __init__(self, parent=None):
        """
        Args:
            parent (optional): Parent object
        """
        super(EditWidget, self).__init__(parent)

        # Get app's global font size
        font = QFont()
        font.setPointSize(AppConfig.get_font_size())
        self.setFont(font)

        self.logger = logging.getLogger('edit_widget')

        self.debug = AppConfig.get_debug()

    def setDocument(self, document):
        # Override setDocument() to allow additional actions like emit the document set signal
        super().setDocument(document)
        # Emit document set signal to notify listeners
        self.content_set.emit()

    def setPlainText(self, text):
        # Override setDocument() to allow additional actions like emit the document set signal
        super().setPlainText(text)
        # Emit document set signal to notify listeners
        self.content_set.emit()

    def find_block_by_number(self, block_number: int) -> Union[QTextBlock, None]:
        """
        Notice: Very resource greedy method, avoid it, refactor or change the logic.

        Args:
            block_number (int): Block number to find

        Returns:
            Union[QTextBlock, None]: Returns either a block found by number or None
        """
        cursor = QTextCursor(self.document())
        cursor.movePosition(QTextCursor.MoveOperation.Start)

        while not (cursor.atEnd()
                   # Sometimes, cursor.atEnd() doesn't work, check it
                   or cursor.block().blockNumber() == self.document().blockCount() - 1):
            current_block = cursor.block()
            user_data = current_block.userData()  # type: Union[QTextBlockUserData, TextBlockData]

            if user_data and user_data.block_number == block_number:
                return current_block

            cursor.movePosition(QTextCursor.MoveOperation.NextBlock)

        return None

    def get_current_block(self) -> QTextBlock:
        return self.textCursor().block()

    def keyPressEvent(self, event):
        """
        This method handle key press events.

        Args:
            event (QKeyEvent): Event object
        """
        if event.key() == Qt.Key.Key_Backtab:
            """
            Shift + Tab action to shift the text block left either on a one tab or 4 spaces.
            """
            self.process_back_tab()
        else:
            return QPlainTextEdit.keyPressEvent(self, event)

    def process_back_tab(self) -> None:
        """
        Shift the text block left either on a one tab or 4 spaces.
        """
        find_cursor = self.textCursor()
        """
        Current position and selection
        pos - is the position right after the char located behind the cursor; end of selection if selected
        anchor - either the beginning of the selection or the same as pos
        """
        pos = find_cursor.position()
        anchor = find_cursor.anchor()  # Where a selection starts (can be the same as above)

        # Start batch editing
        find_cursor.beginEditBlock()

        # QTextCursor.MoveMode.MoveAnchor applied by default
        find_cursor.setPosition(pos)

        # Count the number of spaces to reduce
        spaces_cnt = 0

        # Start moving the cursor position one symbol backward (QTextCursor.MoveOperation.Left)
        find_cursor.setPosition(pos - 1, QTextCursor.MoveMode.KeepAnchor)
        # Get the character on that new position
        curr_char = find_cursor.document().characterAt(find_cursor.position())
        if curr_char == " ":
            spaces_cnt += 1

        # If non-space characters continue moving cursor to any space alike position
        while (not find_cursor.document().characterAt(find_cursor.position()).isspace()
               and find_cursor.positionInBlock() > 0):
            """
            KeepAnchor makes text selected if combine it with: self.setTextCursor(find_cursor)
            """
            find_cursor.movePosition(QTextCursor.MoveOperation.Left, QTextCursor.MoveMode.MoveAnchor, 1)
            if not find_cursor.document().characterAt(find_cursor.position()).isspace():
                self.setTextCursor(find_cursor)

        # if the cursor text is a space
        if str(find_cursor.selectedText()) == " ":
            # Continue while space characters and not at beginning of the line or when a limit of spaces reached
            while (str(find_cursor.selectedText()).strip() == ""
                   and find_cursor.positionInBlock() > 0
                   and spaces_cnt < 4):
                # Move cursor to the left
                find_cursor.movePosition(QTextCursor.MoveOperation.Left, QTextCursor.MoveMode.KeepAnchor, 1)
                self.setTextCursor(find_cursor)
                cursor_char = find_cursor.document().characterAt(find_cursor.position())
                # Non-space symbol
                if cursor_char != " ":
                    """
                    " "  – Space
                    "\t" – Horizontal tab
                    "\v" – Vertical tab
                    "\n" – Newline
                    "\r" – Carriage return
                    "\f" – Feed
                    """
                    # Non-space character
                    if not cursor_char.isspace():
                        # Move anchor to a new non-space position
                        find_cursor.setPosition(find_cursor.position() + 1, QTextCursor.MoveMode.MoveAnchor)
                        # Move cursor to a new position before the non-space character
                        self.setTextCursor(find_cursor)
                        # Do not process spaces
                        spaces_cnt = 0
                    elif cursor_char == "\t":
                        # Keep anchor to add the preceding spaces to selection
                        find_cursor.setPosition(find_cursor.position() + 1, QTextCursor.MoveMode.KeepAnchor)
                        # Move cursor to a new position before the tab
                        self.setTextCursor(find_cursor)
                    break
                # Count a spaces to reduce
                spaces_cnt += 1

        if spaces_cnt > 0:
            # Reduce the spaces if any exist
            find_cursor.removeSelectedText()
        elif str(find_cursor.selectedText()) == "\t":
            # The prior character is a tab, remove it
            find_cursor.removeSelectedText()
            # Move cursor further left
            find_cursor.setPosition(anchor - 1)
            # Keep anchor from previous setPosition() and continue to move further left
            find_cursor.setPosition(pos - 1, QTextCursor.MoveMode.KeepAnchor)
        else:
            # Restore position
            find_cursor.setPosition(anchor)
            find_cursor.setPosition(pos, QTextCursor.MoveMode.KeepAnchor)

        # Stop batch editing
        find_cursor.endEditBlock()

    def contextMenuEvent(self, event):
        # Create the standard context menu
        menu = self.createStandardContextMenu()

        # Modify the menu font to hint the size
        menu.setFont(self.font())

        """
        # Optionally add new actions or modify existing ones
        customAction = menu.addAction("Custom Action")
        customAction.triggered.connect(self.custom_action_triggered)
        """

        # Display the menu
        menu.exec_(event.globalPos())



import datetime
import logging
import json
import re
import os

from typing import Any, Union, Iterable
from json import JSONDecodeError

from .app_config import AppConfig
from .helpers.file_helper import read_file
from .encrypt.enc_helper import EncHelper

from .exceptions.file_header_empty_exception import FileHeaderEmptyException


class FileHeader:
    """
    File header stores the data about the file and its contents.
    Header is a line of data surrounded by HTML-comment tags to avoid having it visible on external resources.

    To avoid hassle the header is only appears in files created within the Notolog editor.
    Anyway, a header is required for encrypted files as it contains such important info as salt or other file data.
    """

    HEADER_TPL = '<!-- %s -->'

    def __init__(self):
        super(FileHeader, self).__init__()

        self.logger = logging.getLogger('header')

        self.logging = AppConfig.get_logging()
        self.debug = AppConfig.get_debug()

        if self.debug:
            self.logger.debug('Header helper is engaged')

        self.header = None

    def get_new(self, is_enc: bool = False):
        if not self.header:
            self.header = self.generate(is_enc=is_enc)
        else:
            if self.logging:
                self.logger.warning('Attempt to create new file header on top of existing one!')
        return self

    def generate(self, is_enc: bool = False) -> json:
        created = updated = datetime.datetime.now()
        header = {'notolog.app': {'created': created, 'updated': updated}}
        if is_enc:
            header['notolog.app'].update(self.generate_enc())
        return header

    def generate_enc(self) -> json:
        return {'enc': {'slt': EncHelper.generate_salt(), 'itr': EncHelper.get_default_iterations(), 'hint': ''}}

    def set_encrypted(self):
        self.header['notolog.app'].update(self.generate_enc())

    def is_valid(self) -> bool:
        return (self.header is not None
                and isinstance(self.header, Iterable)
                and 'notolog.app' in self.header)

    def is_file_encrypted(self) -> bool:
        if not self.is_valid():
            return False
        _header = self.header['notolog.app']
        return 'enc' in _header and isinstance(_header['enc'], dict)

    def validate_enc(self) -> None:
        """
        Ensure that the header is valid and do migrations if necessary.
        """
        pass

    def refresh(self) -> None:
        date_now = datetime.datetime.now()
        # Update only existing header
        if self.is_valid():
            self.header['notolog.app']['updated'] = date_now
            if 'created' not in self.header['notolog.app'] or self.header['notolog.app']['created'] is None:
                self.header['notolog.app']['created'] = date_now
        else:
            if self.debug:
                self.logger.debug('File header is empty')

    def get_param(self, param: Any, default: Any = None) -> Union[str, dict, None]:
        """
        Get header's param by name.
        @param param: name of the param
        @param default: default value if param is not set or not found
        @return: either string or None
        """
        return (self.header['notolog.app'][param]
                if self.is_valid() and param in self.header['notolog.app']
                else default)

    def set_param(self, param: str, value: str) -> None:
        """
        Set params to the header.
        @param param: name of the param
        @param value: value of the param
        @return: None
        """
        if self.is_valid():
            self.header['notolog.app'][param] = value
        else:
            raise FileHeaderEmptyException()

    def get_enc_param(self, param: Any) -> Union[str, None]:
        """
        Get encryption related header's param by name.
        @param param: name of the param
        @return: either string or None
        """
        enc = self.get_param('enc')
        if enc and param in enc:
            return enc[param]
        return None

    def set_enc_param(self, param: str, value: str) -> None:
        """
        Set encryption params to the header.
        @param param: name of the param
        @param value: value of the param
        @return: None
        """
        enc = self.get_param('enc')
        if enc:
            enc.update({param: value})
            self.header['notolog.app'].update({'enc': enc})
        else:
            raise FileHeaderEmptyException()

    def pack(self, content: str = None, encode: bool = False) -> str:
        header_line = repr(self)
        delimiter = '\n'

        if encode:
            delimiter = delimiter.encode("utf-8")
            if header_line and type(header_line) is not bytes:
                header_line = header_line.encode("utf-8")
            if content is not None and type(content) is not bytes:
                content = content.encode("utf-8")

        if header_line and content is not None:
            # Return header and content, even if the content is empty
            return header_line + delimiter + content
        elif content is not None:
            if self.logging:
                self.logger.warning(f'File content exist but header is None {self}')
            # Return content without header
            return content
        else:
            if self.logging:
                self.logger.warning(f'File header exist but content is None {self}')
            # Return header without content
            return header_line

    def load_file(self, file_path: str) -> tuple[Any, str]:
        file_data = read_file(file_path)
        return self.load(file_data)

    def load(self, file_data: str) -> tuple[Any, Union[str, None]]:
        try:
            file_header_line = file_data.splitlines()[0]
        except (IndexError, AttributeError):
            if self.debug:
                self.logger.debug('File header is not found')
            return self, file_data

        self.header = None
        try:
            search = re.search(self.HEADER_TPL % '(.*?)', file_header_line, re.IGNORECASE)
            file_header_json = search.group(1) if search else None
            self.header = json.loads(file_header_json)
            # Run migrations here if needed
            self.validate_enc()
        except (TypeError, JSONDecodeError):
            if self.debug:
                self.logger.debug('File header is empty')
            return self, file_data

        if self.is_valid():
            # Splitting file_data into lines
            lines = file_data.splitlines()
            if len(lines) >= 2:
                file_body = "\n".join(lines[1:])
            else:
                file_body = None
            return self, file_body

        if self.debug:
            self.logger.debug('File header is empty')  # Suppose to be a valid situation if not set yet or not created

        return self, file_data

    def read(self, file_path: str) -> Union[str, None]:
        if os.path.isfile(file_path):
            with open(file_path, encoding='utf-8') as f:
                header_line = f.readline().strip('\n')
                return header_line
        else:
            if self.logging:
                self.logger.warning('File not found "%s"' % file_path)
        return None

    def __repr__(self) -> str:
        """
        Return header data as a formatted string
        """

        if self.header:
            return self.HEADER_TPL % json.dumps(self.header, default=str)

        return ''



# Spanish lexemes settings_dialog.py
lexemes = {
    # Configuración
    "window_title": "Configuración",

    "button_close": "Cerrar",

    "tab_general": "General",
    "tab_editor_config": "Editor",
    "tab_viewer_config": "Visor",
    "tab_ai_config": "Configuración IA",

    "general_app_config_label": "Configuración de la app",
    "general_app_language_label": "Idioma",
    "general_app_language_combo_placeholder_text": "Elegir un idioma",
    "general_app_language_combo_accessible_description": "Idioma de la interfaz de la app",
    "general_app_theme_label": "Tema",
    "general_app_theme_combo_placeholder_text": "Elegir un tema",
    "general_app_theme_combo_accessible_description": "Tema de la interfaz de la app",
    "general_app_main_menu_label": "Menú Principal",
    "general_app_main_menu_checkbox": "Mostrar menú principal",
    "general_app_main_menu_checkbox_accessible_description": "Mostrar el menú desplegable principal de la aplicación",
    "general_app_font_size_label": "Tamaño de fuente: {size}pt",
    "general_app_font_size_slider_accessible_description": "Ajustar el tamaño de fuente global de la aplicación",

    "general_statusbar_label": "Barra de estado",
    "general_statusbar_show_global_cursor_position_checkbox": "Mostrar la posición global del cursor",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Mostrar la posición global del cursor en la barra de estado",

    "editor_config_label": "Configuración del editor",
    "editor_config_show_line_numbers_checkbox": "Mostrar números de línea",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Mostrar números de línea en el editor",

    "viewer_config_label": "Configuración del visor",
    "viewer_config_process_emojis_checkbox": "Convertir emojis de texto en gráficos",
    "viewer_config_process_emojis_checkbox_accessible_description":
        "Convertir emojis de texto en representaciones gráficas",
    "viewer_config_highlight_todos_checkbox": "Resaltar TODOs",
    "viewer_config_highlight_todos_checkbox_accessible_description": "Resaltar etiquetas TODO en el texto",
    "viewer_config_open_link_confirmation_checkbox": "Requerir confirmación para abrir enlaces",
    "viewer_config_open_link_confirmation_checkbox_accessible_description":
        "Solicitar confirmación antes de abrir enlaces",
    "viewer_config_save_resources_checkbox": "Guardar automáticamente imágenes externas en el disco",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Guarda automáticamente copias de imágenes externas en el disco para acceso sin conexión.",

    "ai_config_openai_api_label": "API de OpenAI",
    "ai_config_openai_api_url_input_placeholder_text": "URL de la API",
    "ai_config_openai_api_url_input_accessible_description": "URL de la API de OpenAI",
    "ai_config_openai_api_key_input_placeholder_text": "Clave de la API",
    "ai_config_openai_api_key_input_accessible_description": "Clave de la API de OpenAI",
    "ai_config_openai_api_supported_models_label": "Modelos soportados",
    "ai_config_ai_model_names_combo_placeholder_text": "Elegir modelo",
    "ai_config_ai_model_names_combo_accessible_description": "Modelos soportados para elegir",

    "ai_config_base_label": "Parámetros básicos",
    "ai_config_base_system_prompt_label": "Prompt del sistema",
    "ai_config_base_system_prompt_edit_placeholder_text": "Prompt del sistema base que precede a cada solicitud",
    "ai_config_base_system_prompt_edit_accessible_description":
        "Prompt del sistema base que precede a cada solicitud. Texto plano.",
    "ai_config_base_response_max_tokens_label": "Máximo de tokens de respuesta",
    "ai_config_base_response_max_tokens_input_accessible_description":
        "Número máximo de tokens para recibir en la respuesta",
}


"""
AI Assistant Dialog Class

This class is designed to facilitate and manage interactions between the user and various external plugins or APIs.
It acts as a bridge to send queries to and receive responses from selected plugins, ensuring seamless integration and
efficient data handling. The class leverages user inputs and contextual information to interact with APIs dynamically.

Features:
- Sends requests and receives responses based on user inputs or context.
- Supports robust error handling to manage API limitations or failures, ensuring consistent application performance.
- Provides functionality that can be adjusted based on the app
"""

from PySide6.QtCore import Qt, QUrl, QByteArray, QSize
from PySide6.QtNetwork import QNetworkRequest, QNetworkAccessManager, QNetworkReply
from PySide6.QtWidgets import QDialog, QVBoxLayout, QWidget, QLineEdit, QTextEdit, QPushButton
from PySide6.QtWidgets import QLabel, QSizePolicy, QHBoxLayout
from PySide6.QtGui import QPixmap

import json
import logging

from . import Settings
from . import AppConfig
from . import Lexemes
from . import ThemeHelper

from .vertical_line_spacer import VerticalLineSpacer

from ..ui.rotating_label import RotatingLabel
from ..enums.ai_model_names import AiModelNames


class AIAssistant(QDialog):

    def __init__(self, parent):
        super().__init__(parent, Qt.WindowType.Dialog)

        self.parent = parent

        # Apply font from the dialog instance to the label
        self.setFont(self.parent.font())

        self.logger = logging.getLogger('ai_assistant')

        self.logging = AppConfig.get_logging()
        self.debug = AppConfig.get_debug()

        self.settings = Settings(parent=self)

        # TODO add more AI providers and a setting of which one is in use
        self.api_url = self.settings.ai_config_openai_url
        self.api_key = self.settings.ai_config_openai_key
        self.inference_model = AiModelNames.__getitem__(self.settings.ai_config_openai_model).value
        self.prompt_system = self.settings.ai_config_base_system_prompt
        self.prompt_user = "%s"
        self.response_max_tokens = self.settings.ai_config_base_response_max_tokens

        self.theme_helper = ThemeHelper()

        # Default language setup, change to settings value to modify it via UI
        self.lexemes = Lexemes(self.settings.app_language, default_scope='ai_assistant')

        self.layout = QVBoxLayout(self)

        self.setWindowTitle(self.lexemes.get('dialog_title'))

        # Set dialog size derived from the main window size
        main_window_size = self.parent.size()
        dialog_width = int(main_window_size.width() * 0.5)
        dialog_height = int(main_window_size.height() * 0.5)
        # self.setMinimumSize(dialog_width, dialog_height)
        self.resize(dialog_width, dialog_height)

        self.setStyleSheet(self.theme_helper.get_css('ai_assistant'))

        # Prompt input field
        self.prompt_input = QLineEdit(self)
        self.prompt_input.sizeHint()
        self.prompt_input.setPlaceholderText(
            self.lexemes.get('dialog_prompt_input_placeholder_text'))
        self.prompt_input.setAccessibleDescription(
            self.lexemes.get('dialog_prompt_input_accessible_description'))
        self.prompt_input.returnPressed.connect(self.send_request)
        self.layout.addWidget(self.prompt_input)

        # Text output field for displaying JSON response
        self.response_output = QTextEdit(self)
        self.response_output.sizeHint()
        self.response_output.setReadOnly(True)
        self.response_output.setPlaceholderText(
            self.lexemes.get('dialog_response_output_placeholder_text'))
        self.response_output.setAccessibleDescription(
            self.lexemes.get('dialog_response_output_accessible_description'))
        self.layout.addWidget(self.response_output)

        label_size = QSize(48, 48)
        label_pixmap = QPixmap(self.theme_helper.get_icon(theme_icon='arrow-repeat.svg').pixmap(label_size))
        # Use custom Pixmap class to asure the transformation (rotation)
        self.background_label = RotatingLabel(pixmap=label_pixmap, parent=self)
        self.background_label.resize(label_size)
        # self.background_label.setPixmap(label_pixmap)
        self.background_label.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)
        # Set at the center of the result text field
        self.background_label.setGeometry(self.width() // 2 - self.background_label.width() // 2,
                                          self.height() // 2 - self.background_label.height() // 2,
                                          self.background_label.width(),
                                          self.background_label.height())
        # Rise it up on the first plane
        self.background_label.raise_()
        # Hide on the very beginning as it will be set to visible upon request
        self.background_label.hide()

        status_bar_widget = QWidget(self)
        status_bar_layout = QHBoxLayout(status_bar_widget)

        model_label = QLabel(text=self.lexemes.get('dialog_usage_model_label'))
        model_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
        model_label.sizeHint()
        model_label.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        model_label.setObjectName("ai_assistant_dialog_usage_model_label")
        # model_label.setStyleSheet("QLabel {color: grey;}")
        status_bar_layout.addWidget(model_label)

        status_bar_layout.addWidget(VerticalLineSpacer())

        self.model_label = QLabel(text=self.inference_model)
        self.model_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.model_label.sizeHint()
        self.model_label.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        self.model_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
        status_bar_layout.addWidget(self.model_label)

        central_spacer = QWidget()
        central_spacer.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        status_bar_layout.addWidget(central_spacer)

        tokens_label = QLabel(text=self.lexemes.get('dialog_usage_tokens_label'))
        tokens_label.setAlignment(Qt.AlignmentFlag.AlignRight)
        tokens_label.sizeHint()
        tokens_label.setObjectName("ai_assistant_dialog_usage_tokens_label")
        # tokens_label.setStyleSheet("QLabel {color: grey;}")
        status_bar_layout.addWidget(tokens_label)
        status_bar_layout.addWidget(VerticalLineSpacer(), alignment=Qt.AlignmentFlag.AlignRight)
        self.tokens_prompt_label = QLabel()
        self.tokens_prompt_label.setAlignment(Qt.AlignmentFlag.AlignRight)
        self.tokens_prompt_label.sizeHint()
        status_bar_layout.addWidget(self.tokens_prompt_label)
        status_bar_layout.addWidget(VerticalLineSpacer(), alignment=Qt.AlignmentFlag.AlignRight)
        self.tokens_answer_label = QLabel()
        self.tokens_answer_label.setAlignment(Qt.AlignmentFlag.AlignRight)
        self.tokens_answer_label.sizeHint()
        status_bar_layout.addWidget(self.tokens_answer_label)
        status_bar_layout.addWidget(VerticalLineSpacer(), alignment=Qt.AlignmentFlag.AlignRight)
        self.tokens_total_label = QLabel()
        self.tokens_total_label.setAlignment(Qt.AlignmentFlag.AlignRight)
        self.tokens_total_label.sizeHint()
        status_bar_layout.addWidget(self.tokens_total_label)

        self.layout.addWidget(status_bar_widget)
        # Update usage with initial params
        self.update_usage()

        # Submit button
        self.send_button = QPushButton(self.lexemes.get('dialog_button_send_request'), self)
        self.send_button.setFont(self.font())
        self.layout.addWidget(self.send_button)
        self.send_button.clicked.connect(self.send_request)

        # Set the main layout for the dialog
        self.setLayout(self.layout)

    def handle_response(self, reply: QNetworkReply, error_code=None):
        # Get received status code, say 200
        status_code = reply.attribute(QNetworkRequest.Attribute.HttpStatusCodeAttribute)

        if self.debug:
            self.logger.debug(f'AI assistant response: {reply}, {error_code}, {status_code}')

        # Return the QNetworkReply object that emitted the signal.
        # For example: NetworkError.UnknownNetworkError
        # reply = self.sender()  # type: QNetworkReply

        error = None
        # Make sure there is no error
        if reply.error() == QNetworkReply.NetworkError.NoError:
            # The request was successful
            data = reply.readAll()  # type: QByteArray

            if self.debug:
                self.logger.debug(f'Raw RESPONSE [{status_code}]: {data}')

            reply.deleteLater()  # Clean up the QNetworkReply object

            # json_document = QJsonDocument.fromJson(data)
            # json_data = json_document.object()

            # Convert QByteArray to string
            res_string = data.toStdString()  # Or: str(data.data(), encoding='utf-8')
            if self.debug:
                self.logger.debug(f'Result multi-line STRING: {res_string}')

            # Clean up the string and make one line
            json_str = ''.join(line.strip() for line in res_string.splitlines())

            if self.debug:
                self.logger.debug(f'Result STRING: {json_str}')

            result_message = self.lexemes.get('network_connection_error_empty', scope='common')
            # Parse JSON response
            try:
                json_data = json.loads(json_str)
                if self.debug:
                    self.logger.debug(f"Result JSON: {json_data}")
                if ('choices' in json_data
                        and len(json_data['choices']) > 0
                        # Legacy completions
                        # and 'text' in json_data['choices'][0]):
                        and 'message' in json_data['choices'][0]
                        and 'content' in json_data['choices'][0]['message']):
                    # Legacy completions
                    # self.response_output.setPlainText(str(json_data['choices'][0]['text']).strip())
                    self.response_output.setPlainText(str(json_data['choices'][0]['message']['content']).strip())
                if 'usage' in json_data:
                    self.update_usage(json_data['usage'])
                if 'model' in json_data:
                    if hasattr(self, 'model_label'):
                        self.model_label.setText(json_data['model'] if len(json_data['model']) > 0 else '?')
            except json.JSONDecodeError as e:
                if self.logging:
                    self.logger.warning("Error decoding JSON: %s" % e)
        elif reply.error() == QNetworkReply.NetworkError.HostNotFoundError:
            # The host was not found, indicating possible DNS issues or no internet connection
            result_message = self.lexemes.get('network_connection_error_connection_or_dns', scope='common')
        elif reply.error() == QNetworkReply.NetworkError.ConnectionRefusedError:
            # Connection was refused, indicating the server might be down or there are network issues
            result_message = self.lexemes.get('network_connection_error_connection_refused', scope='common')
        elif reply.error() == QNetworkReply.NetworkError.TimeoutError:
            # The connection timed out, indicating network issues
            result_message = self.lexemes.get('network_connection_error_connection_timed_out', scope='common')
        elif reply.error() == QNetworkReply.NetworkError.ContentNotFoundError:
            # The requested page or resource is not found
            result_message = self.lexemes.get('network_connection_error_connection_404_error', scope='common')
        else:
            # Handle other errors
            result_message = self.lexemes.get('network_connection_error_generic_with_status_code', scope='common',
                                              status_code=status_code)

        if error_code is not None:
            if self.logging:
                self.logger.warning(result_message)
                self.logger.warning(f"Failed to fetch update information: {reply.errorString()}")
            # Set up response status visible within response field
            self.response_output.setPlainText(result_message)

        self.set_status_ready()

    def update_usage(self, usage: dict = None):
        if hasattr(self, 'tokens_prompt_label'):
            prompt_tokens_cnt = (usage['prompt_tokens'] if usage and 'prompt_tokens' in usage else 0)
            self.tokens_prompt_label.setText(self.lexemes.get('dialog_usage_tokens_prompt',
                                                              tokens=prompt_tokens_cnt))
        if hasattr(self, 'tokens_answer_label'):
            answer_tokens_cnt = (usage['completion_tokens'] if usage and 'completion_tokens' in usage else 0)
            self.tokens_answer_label.setText(self.lexemes.get('dialog_usage_tokens_prompt',
                                                              tokens=answer_tokens_cnt))
        if hasattr(self, 'tokens_total_label'):
            total_tokens_cnt = (usage['total_tokens'] if usage and 'total_tokens' in usage else 0)
            self.tokens_total_label.setText(self.lexemes.get('dialog_usage_tokens_total',
                                                             tokens=total_tokens_cnt))

    def send_request(self) -> None:
        # Get context for the prompt
        prompt_context = self.prompt_input.text()
        if len(prompt_context) == 0:
            self.response_output.setText(self.lexemes.get('dialog_response_output_notice_empty_text'))
            return

        url = QUrl(self.api_url)  # API entrypoint
        request = QNetworkRequest(url)

        # Set Authorization header
        access_token = self.api_key
        request.setRawHeader(b"Authorization", bytes("Bearer " + access_token, encoding="utf-8"))

        # Set Content-Type header
        request.setHeader(QNetworkRequest.KnownHeaders.ContentTypeHeader, "application/json")

        # Legacy completions
        """
        Consider these params as well:
        * "logprobs": None,
        * "echo": false,
        * * "stop": null,
        * "presence_penalty": 0,
        * "frequency_penalty": 0,
        * * "best_of": 1,
        * "logit_bias": None,
        * "user": "",
        """

        # New completions
        post_params = {
            "model": self.inference_model,
            "messages": [
                {
                    "role": "system",
                    "content": self.prompt_system,
                },
                {
                    "role": "user",
                    "content": self.prompt_user % prompt_context,
                }
            ],
            "temperature": 0.2,
            "top_p": 1,
            "n": 1,
            "stream": False,
        }
        # If response max tokens set
        if self.response_max_tokens:
            post_params.update({"max_tokens": self.response_max_tokens})

        json_post_params = json.dumps(post_params)

        # Set request data
        request_data = QByteArray(json_post_params.encode("utf-8"))

        # Set request type to POST
        request.setRawHeader(b"Custom-Request", b"POST")

        self.set_status_waiting()

        # Set request data
        manager = QNetworkAccessManager(self)
        reply = manager.post(request, request_data)

        # Connect finished signal
        reply.finished.connect(lambda _reply=reply: self.handle_response(_reply))
        # Connect error signal
        reply.errorOccurred.connect(lambda error, _reply=reply: self.handle_response(_reply, error))

    def set_status_waiting(self):
        # Set loader cursor whilst loading content
        self.setCursor(Qt.CursorShape.WaitCursor)
        # Disable prompt input field
        self.prompt_input.setDisabled(True)
        # Clear response field from prev results if set
        self.response_output.clear()
        # Disconnect Enter action onto the input field
        self.prompt_input.returnPressed.disconnect()
        # Disable send request button
        self.send_button.setDisabled(True)
        # Show progress label
        self.background_label.show()

    def set_status_ready(self):
        # Set cursor back
        self.setCursor(Qt.CursorShape.ArrowCursor)
        # Enable prompt input field
        self.prompt_input.setEnabled(True)
        # Connect Enter action onto the input field
        self.prompt_input.returnPressed.connect(self.send_request)
        # Enable send request button
        self.send_button.setEnabled(True)
        # Hide progress label
        self.background_label.hide()



from PySide6.QtWidgets import QFrame


class VerticalLineSpacer(QFrame):
    """
    Vertical line separator for the status bar.
    """

    def __init__(self):
        super(VerticalLineSpacer, self).__init__()

        # https://doc.qt.io/qt-6/qframe.html#Shape-enum
        self.setFrameShape(QFrame.Shape.VLine)

        # https://doc.qt.io/qt-6/qframe.html#Shadow-enum
        self.setFrameShadow(QFrame.Shadow.Raised)

        self.setLineWidth(1)



from PySide6.QtCore import Qt, QObject, QUrl
from PySide6.QtWidgets import QTextBrowser
from PySide6.QtGui import QTextCursor, QSyntaxHighlighter, QTextCharFormat

import logging
import re
import base64

from typing import Union

from .settings import Settings
from .app_config import AppConfig
from .highlight.view_highlighter import ViewHighlighter
from .lexemes.lexemes import Lexemes


class ViewProcessor:
    """
    View mode pre-result procession by modifying loaded QTextDocument
    """

    def __init__(self, highlighter: Union[QSyntaxHighlighter, ViewHighlighter]):
        """
        Args:
            highlighter (Union[QSyntaxHighlighter, ViewHighlighter]):
            Highlighter that holds document to apply any modifications.
        """
        self.highlighter = highlighter
        self.doc = self.highlighter.document()

        self.logging = AppConfig.get_logging()
        self.debug = AppConfig.get_debug()

        self.settings = Settings()

        self.logger = logging.getLogger('view_processor')

        # Default language setup, change to settings value to modify it via UI
        self.lexemes = Lexemes(self.settings.app_language, default_scope='common')

        if self.debug:
            self.logger.info('Characters count %d' % self.doc.characterCount())

        self.blocks = []
        self.blocks_start = []
        self.blocks_end = []

        # self.doc.contentsChanged.connect(self.process)

        cursor = QTextCursor(self.doc)
        self.cursor_pos_orig = cursor.position()

    def restore_cursor_pos(self):
        cursor = QTextCursor(self.doc)
        # Restore original cursor position
        cursor.setPosition(self.cursor_pos_orig, QTextCursor.MoveMode.MoveAnchor)
        parent = self.doc.parent()  # type: QObject
        parent_widget = parent.get_view_widget()  # type: QTextBrowser
        parent_widget.setTextCursor(cursor)

    def process(self):
        self.blocks.clear()
        cursor = QTextCursor(self.doc)
        cursor.setPosition(0)

        def process_block(cursor_internal: QTextCursor) -> None:
            """
            Find open/close tokens in block.
            @param cursor_internal: active cursor
            @return: None
            """
            block = cursor_internal.block()
            pattern = r"<details.*?>"
            # Find all matches of the pattern in provided string
            matches = re.finditer(pattern, block.text())
            for match in matches:
                open_start = match.start()
                open_end = match.end()
                open_length = open_end - open_start
                self.blocks_start.append((block.position() + open_start, open_length))
            pattern = r"</details>"
            # Find all matches of the pattern in provided string
            matches = re.finditer(pattern, block.text())
            for match in matches:
                close_start = match.start()
                close_end = match.end()
                close_length = close_end - close_start
                self.blocks_end.append((block.position() + close_start, close_length))

        cursor.movePosition(QTextCursor.MoveOperation.Start)
        # Iterate through each block
        while not (cursor.atEnd()
                   # cursor.atEnd() not enough as the cursor unable to move next block at the very end
                   or (cursor.block().blockNumber() > 0
                       and cursor.block().blockNumber() == self.doc.blockCount() - 1)):
            # Process block
            process_block(cursor)
            # Move to the next block
            cursor.movePosition(QTextCursor.MoveOperation.NextBlock)

            # To avoid never ending cycle when only one block is available
            if self.doc.blockCount() == 1:
                break

        # Process block at the very end
        process_block(cursor)

        res = []
        self.blocks_start.reverse()
        # Enumerate by end token to find the closest start token
        for index, value in enumerate(self.blocks_end):
            close_start, close_length = value
            if index not in res:
                """
                Extra data allows independent processing of tokens and their content:
                o - open, start position
                ol - open length, start token length, say <details> length is 9
                c - close, close position
                cl - close length, close token length (the same approach as for open length)
                """
                res.insert(index, {'o': None, 'ol': None, 'c': close_start, 'cl': close_length, 'g': None, 'l': 0})
            # Enumerate start token list backward unsetting matched element
            for _index, _value in enumerate(self.blocks_start):
                open_start, open_length = _value
                if _value < value:
                    res[index].update({'o': open_start, 'ol': open_length})
                    del self.blocks_start[_index]
                    break

        if self.debug:
            self.logger.debug('Unsorted open-close mapping %s' % res)

        self.blocks = res

        if self.blocks:
            self.sort_blocks()
            self.collapse_blocks()

        # Restore original cursor position
        self.restore_cursor_pos()

    def sort_blocks(self):
        if self.debug:
            self.logger.debug('Sorting elements by group and nesting level...')

        self.blocks.sort(key=lambda x: x['o'] if 'o' in x and x['o'] is not None else 0)

        if self.debug:
            self.logger.info('Sorted result of the open-close mapping %s' % self.blocks)

        group = 0
        for i, _data in enumerate(self.blocks):
            if not _data['l']:
                # Update group only for each root element
                group = i
                self.blocks[i].update({'g': group})
            # Iterate through remaining elements to find possible nesting
            for j, _other_data in enumerate(self.blocks[i + 1:], start=i + 1):
                if _data['o'] is None or _other_data['o'] is None or _data['c'] is None or _other_data['c'] is None:
                    # TODO show error sign at status bar
                    return
                # Check if other element is nested within the current element
                if _data['o'] < _other_data['o'] and _data['c'] > _other_data['c']:
                    self.blocks[j].update({'l': (_data['l'] + 1 if 'l' in _data else 1), 'g': group})

        """
        Sort elements by:
        1. By group first (nested elements or singles)
        2. Nesting level (more means deeper nesting)
        3. Open token position (DESC order)
        """
        self.blocks.sort(key=lambda x: (-x['g'] if x['g'] is not None else 0,
                                        -x['l'] if x['l'] is not None else 0,
                                        -x['o'] if x['o'] is not None else 0))  # reverse=True

        if self.debug:
            self.logger.info('Final result of the open-close mapping with nesting value %s' % self.blocks)

    def collapse_blocks(self):
        cursor = QTextCursor(self.doc)
        # Keep prev states to check within a loop
        prev_group = None
        prev_level = 0
        replacement_text_len = 0
        replacement_text_lengths = {}
        for pair in self.blocks:
            """
            Blocks being processed in a particular order.
            The sorting applied at the end of the sort_blocks() method.
            """
            pos_group = pair['g']
            pos_level = pair['l']
            # When start processing another group of nested (or single) elements
            if pos_group != prev_group:
                prev_group = pos_group
                replacement_text_len = 0
                replacement_text_lengths = {}
            # Previous processed pair['l'] value have to be set
            if prev_level not in replacement_text_lengths:
                replacement_text_lengths[prev_level] = 0
            """
            Text has been replaced with a new one, different length. Include it in length calculation.
            The value is from the previous iteration but can be modified at this one if the conditions match.
            """
            replacement_text_lengths[prev_level] += replacement_text_len
            # If the previous nesting level is the same as a current one reset the length correction then
            if pos_level == prev_level:
                replacement_text_len = 0

            if pair['o'] is None or pair['c'] is None:
                if self.debug:
                    self.logger.warning('Error occurred! Incomplete open/close tag data %s' % pair)
                return

            """
            The most nested blocks being processed first, thus the close token correction,
            as the open token doesn't change its position.
            """
            pos_open = pair['o']
            pos_close = (pair['c'] + pair['cl']  # text + closing token together to get the very end of the text
                         # Combined length correction of all nested elements inside (previous level), say each ones:
                         # [0]...[1][2][/2][/1]...[1][/1]...[/0]
                         + (replacement_text_lengths[prev_level] if pos_level == 0 else replacement_text_len))

            if self.debug:
                self.logger.debug('Cursor position open: %d, close: %d, level: %d ' % (pos_open, pos_close, pos_level))

            cursor.setPosition(pos_open)
            cursor.setPosition(pos_close, QTextCursor.MoveMode.KeepAnchor)
            selected_text = cursor.selectedText()
            if selected_text:
                if self.debug:
                    self.logger.debug('Block text[%d][%d]: %s' % (pos_group, pos_level, selected_text))

                cursor.removeSelectedText()

                """
                Select extra space characters before and after to strip the cursor's text the way where the summary's
                content starts right after the details tag.
                """
                pattern = r"[\s]*?<summary.*?>(.*?)<\/summary>[\s]*?"
                match = re.search(pattern, selected_text, flags=re.DOTALL | re.MULTILINE | re.UNICODE)
                summary = self.lexemes.get('expandable_block_default_title')
                if match:
                    summary = f"{match.group(1).strip()}"
                    pos_open_summary = match.start()
                    pos_close_summary = match.end()
                    selected_text = selected_text[:pos_open_summary] + selected_text[pos_close_summary:]

                """
                There is a possible situation where the formatted html may become invalid.
                Say, "<p>...\n<details> </p>\n\n<summary>". That's because of markdown lib may process paragraph that way.
                Double check it with with a something like exp below, but keep in mind the open tag is also exists:
                pattern = r"<details.*?>[\s]*?(</p>)"
                """

                selected_text = selected_text.strip()
                encoded_text = base64.b64encode(selected_text.encode('utf-8'))
                """
                Insert as a TEXT here, not as an HTML or tags will be stripped after.
                Place extra spaces on each side to allow smooth anchor detection upon click.
                """
                # Do not use id="..." it makes anchor detection unstable, use data attributes instead
                replacement_text = ('<p class="_ds_expand">'
                                    # <a>...</a> will replaced with table upon click
                                    '<a href="expandable:{}" data-group="{}" data-level="{}">'
                                    '<span class="_ds_expand_pointer">▼</span>&nbsp;{}</a>'
                                    '</p>'
                                    .format(encoded_text.decode('utf-8'), pos_group, pos_level, summary))
                replacement_text_len += len(replacement_text) - (pos_close - pos_open)
                cursor.insertText(replacement_text)
                if self.debug:
                    self.logger.debug('Replacement text: %s' % replacement_text)

                # Save current group and level to check them over again within the next iteration.
                prev_group = pos_group
                prev_level = pos_level

    def mouse_click_event(self, event):
        parent = self.doc.parent()  # type: QObject
        parent_widget = parent.get_view_widget()  # type: QTextBrowser

        # Parent actions, like select text with a cursor
        QTextBrowser.mousePressEvent(parent_widget, event)

        cursor = parent_widget.cursorForPosition(event.pos())

        if event.button() == Qt.MouseButton.LeftButton:
            anchor = parent_widget.anchorAt(event.pos())
            if anchor.startswith('expandable'):
                return self.anchor_click_event(QUrl(anchor), cursor)
            elif anchor.startswith('collapsible'):
                # TODO collapsible block
                if self.debug:
                    self.logger.debug('Collapsible block clicked')
                pass
            else:
                if self.debug:
                    cursor.select(QTextCursor.SelectionType.WordUnderCursor)
                    text = cursor.selectedText()
                    self.logger.debug(f'Click context: {text}')

                cursor.clearSelection()
                parent_widget.setTextCursor(cursor)

    def anchor_click_event(self, url: QUrl, cursor: QTextCursor):
        if self.debug:
            # cursor.charFormat().anchorHref()
            self.logger.debug('Anchor clicked, scheme: %s, path: %s' % (url.scheme(), url.path()))

        # This may cause left-right click issue when either the start or end position of token is clicked
        # cursor.movePosition(QTextCursor.MoveOperation.PreviousCharacter, QTextCursor.MoveMode.KeepAnchor)

        cursor.select(QTextCursor.SelectionType.WordUnderCursor)

        if self.debug:
            self.logger.debug('Cursor data, pos: %d, anchor: %s, text: %s'
                              % (cursor.position(), cursor.anchor(), cursor.selectedText()))

        i = 0
        # Move to the anchor caption's start position
        while cursor.charFormat().isAnchor() and (i := i + 1):
            cursor.movePosition(QTextCursor.MoveOperation.PreviousCharacter, QTextCursor.MoveMode.MoveAnchor)
        # Move to the anchor caption's end position
        while i > 0 or cursor.charFormat().isAnchor():
            i -= 1
            cursor.movePosition(QTextCursor.MoveOperation.NextCharacter, QTextCursor.MoveMode.KeepAnchor)
            # Check it after, not in loop
            if cursor.atBlockEnd():
                break

        anchor_label = cursor.selectedText()
        if self.debug:
            self.logger.debug('Anchor text: %s' % anchor_label)

        cursor.removeSelectedText()
        to_decode = url.path()
        if self.debug:
            self.logger.debug('Text to decode: {%s}', to_decode)

        decoded_text = (base64.b64decode(to_decode.encode('utf-8'))
                        .decode('utf-8')
                        .strip())

        if self.debug:
            self.logger.debug('Decoded text: {%s}' % decoded_text)

        # Set default format for inserted text
        default_format = QTextCharFormat()
        cursor.setCharFormat(default_format)
        """
        Insert decoded text, new line either "\n" or "<br/>" depending on insertText() or insertHtml() correspondingly.
        Allow extra space as space symbols could be stripped during conversion.
        """
        cursor.insertHtml('<table class="_n_details">'
                          '<tr><td class="_n_details_summary"><span class="_ds_collapse">{}</span></td></tr>'
                          '<tr><td class="_n_details_content">{}</td></tr>'
                          '</table>'
                          # TODO make it collapsible with '▲'
                          .format(anchor_label.replace('▼', '≡').strip(), decoded_text))



from PySide6.QtCore import Qt
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QStatusBar, QWidget, QLabel, QPushButton, QSizePolicy

import logging
from typing import Union

from . import Settings
from . import AppConfig
from . import Lexemes
from . import ThemeHelper

from .sort_filter_proxy_model import SortFilterProxyModel
from .vertical_line_spacer import VerticalLineSpacer


class StatusBar(QStatusBar):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.parent = parent

        if self.parent and hasattr(self.parent, 'font'):
            # Apply font from the dialog instance to the label
            self.setFont(self.parent.font())

        self.logger = logging.getLogger('statusbar')

        self.logging = AppConfig.get_logging()
        self.debug = AppConfig.get_debug()

        self.settings = Settings(parent=self)

        self.theme_helper = ThemeHelper()

        # Default language setup, change to settings value to modify it via UI
        self.lexemes = Lexemes(self.settings.app_language, default_scope='statusbar')

        self._elements = {}  # Label storage

        self.litter_bin_label = None  # type: Union[QPushButton, None]

        self.data_size_label = None  # type: Union[QLabel, None]
        self.mode_label = None  # type: Union[QLabel, None]
        self.save_progress_label = None  # type: Union[QLabel, None]
        self.encryption_label = None  # type: Union[QLabel, None]
        self.source_label = None  # type: Union[QLabel, None]
        self.cursor_label = None  # type: Union[QLabel, None]

        self.init()

    def init(self):
        # Attach style to the QStatusBar and QLabel
        self.setStyleSheet(self.theme_helper.get_css('statusbar'))

        # Litter bin button to show/hide deleted files
        self.litter_bin_label = QPushButton(self)
        self.litter_bin_label.setCursor(Qt.CursorShape.PointingHandCursor)
        self.litter_bin_label.setFont(self.font())
        self.litter_bin_label.setObjectName('statusbar_litter_bin_label')
        self.litter_bin_label.setFlat(True)
        # Theme icon with a fallback to a system one if needed
        # self.theme_helper.get_color('statusbar_...')
        icon = self.theme_helper.get_icon(theme_icon='trash3.svg',
                                          color=QColor(self.theme_helper.get_color('statusbar_litter_bin_icon_color')))
        self.litter_bin_label.setIcon(icon)
        self.litter_bin_label.setText(self.lexemes.get('statusbar_litter_bin_label'))
        self.litter_bin_label.setToolTip(self.lexemes.get('statusbar_litter_bin_label'))
        self.litter_bin_label.setAccessibleName(self.lexemes.get('statusbar_litter_bin_accessible_name'))
        self.litter_bin_label.clicked.connect(lambda: self.toggle_litter_bin_button())

        self.addWidget(self.litter_bin_label)

        central_spacer = QWidget(self)
        central_spacer.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.addPermanentWidget(central_spacer)

        self.data_size_label = QLabel(self)
        self.data_size_label.setFont(self.font())
        self.addPermanentWidget(self.data_size_label)

        self.save_progress_label = QLabel(self)
        self.save_progress_label.setFont(self.font())
        self.save_progress_label.setVisible(False)
        self.save_progress_label.setText(self.lexemes.get('statusbar_save_progress_label'))
        self.addPermanentWidget(self.save_progress_label)

        self.addPermanentWidget(VerticalLineSpacer())

        self.mode_label = QLabel(self)
        self.mode_label.setFont(self.font())
        """
        Do not set object name similar to the lexeme to avoid double update with unprocessed lexeme's placeholders.
        E.g.: self.mode_label.setObjectName('statusbar_mode_label')
        """
        self.addPermanentWidget(self.mode_label)

        self.addPermanentWidget(VerticalLineSpacer())

        self.encryption_label = QLabel(self)
        self.encryption_label.setFont(self.font())
        self.addPermanentWidget(self.encryption_label)

        self.addPermanentWidget(VerticalLineSpacer())

        self.source_label = QLabel(self)
        self.source_label.setFont(self.font())
        self.addPermanentWidget(self.source_label)

        self.addPermanentWidget(VerticalLineSpacer())

        self.cursor_label = QLabel(self)
        self.cursor_label.setFont(self.font())
        self.addPermanentWidget(self.cursor_label)

    def toggle_litter_bin_button(self):
        self.set_litter_bin_visibility(not self.settings.show_deleted_files)

    def set_litter_bin_visibility(self, visible: bool = False):
        if (hasattr(self, 'parent')
                and hasattr(self.parent, 'tree_proxy_model')
                and isinstance(self.parent.tree_proxy_model, SortFilterProxyModel)):
            # Get tree proxy model to obtain the data
            tree_proxy_model = self.parent.tree_proxy_model  # type: SortFilterProxyModel
            if visible:
                # Add 'deleted' extension to the tree
                tree_proxy_model.add_extension('del'),
                self.litter_bin_label.setIcon(self.theme_helper.get_icon(theme_icon='trash3-fill.svg',
                    color=QColor(self.theme_helper.get_color('statusbar_litter_bin_icon_color_active'))))
                self.settings.show_deleted_files = True
            else:
                # Remove 'deleted' extension from the tree
                tree_proxy_model.remove_extension('del'),
                self.litter_bin_label.setIcon(self.theme_helper.get_icon(theme_icon='trash3.svg',
                    color=QColor(self.theme_helper.get_color('statusbar_litter_bin_icon_color'))))
                self.settings.show_deleted_files = False

    def __getitem__(self, name) -> QLabel:
        """
        To get items like:
            self.statusbar['mode_label']
        @param name: string name of the object
        @return: QLabel object
        """
        return self._elements.get(name, f"Label '{name}' not found")

    def __getattr__(self, name) -> QLabel:
        """
        To get items like:
            self.statusbar.mode_label
        @param name: string name of the object
        @return: QLabel object
        """
        return self._elements.get(name, f"Label '{name}' not found")

    def __setattr__(self, name, value):
        if name == '_attributes' or not name.endswith('_label'):
            super().__setattr__(name, value)
        else:
            if (isinstance(value, QLabel)
                    or isinstance(value, QPushButton)
                    or value is None):
                self._elements[name] = value
            elif self.logging:
                self.logger.warning(f'Trying to set object that is not a QLabel type {type(value)}')




# Japanese lexemes settings_dialog.py
lexemes = {
    # 設定
    "window_title": "設定",

    "button_close": "閉じる",

    "tab_general": "一般",
    "tab_editor_config": "エディター",
    "tab_viewer_config": "ビューア",
    "tab_ai_config": "AI 設定",

    "general_app_config_label": "アプリ設定",
    "general_app_language_label": "言語",
    "general_app_language_combo_placeholder_text": "言語を選択",
    "general_app_language_combo_accessible_description": "アプリのインターフェース言語",
    "general_app_theme_label": "テーマ",
    "general_app_theme_combo_placeholder_text": "テーマを選択",
    "general_app_theme_combo_accessible_description": "アプリのインターフェーステーマ",
    "general_app_main_menu_label": "メインメニュー",
    "general_app_main_menu_checkbox": "メインメニューを表示",
    "general_app_main_menu_checkbox_accessible_description": "アプリのメインドロップダウンメニューを表示",
    "general_app_font_size_label": "フォントサイズ：{size}pt",
    "general_app_font_size_slider_accessible_description": "アプリの全体的なフォントサイズを調整する",

    "general_statusbar_label": "ステータスバー",
    "general_statusbar_show_global_cursor_position_checkbox": "グローバルカーソル位置を表示",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "ステータスバーにグローバルカーソル位置を表示する",

    "editor_config_label": "エディタ設定",
    "editor_config_show_line_numbers_checkbox": "行番号を表示",
    "editor_config_show_line_numbers_checkbox_accessible_description": "エディタ内で行番号を表示する",

    "viewer_config_label": "ビューワ設定",
    "viewer_config_process_emojis_checkbox": "テキスト絵文字をグラフィックに変換",
    "viewer_config_process_emojis_checkbox_accessible_description": "テキスト絵文字をグラフィカルな表現に変換する",
    "viewer_config_highlight_todos_checkbox": "TODOをハイライト",
    "viewer_config_highlight_todos_checkbox_accessible_description": "テキスト内のTODOタグを強調表示する",
    "viewer_config_open_link_confirmation_checkbox": "リンク開く前に確認が必要",
    "viewer_config_open_link_confirmation_checkbox_accessible_description": "リンクを開く前に確認を求める",
    "viewer_config_save_resources_checkbox": "外部画像をディスクに自動保存",
    "viewer_config_save_resources_checkbox_accessible_description":
        "オフラインアクセスのために外部画像のコピーをディスクに自動保存します。",

    "ai_config_openai_api_label": "OpenAI API",
    "ai_config_openai_api_url_input_placeholder_text": "API URL",
    "ai_config_openai_api_url_input_accessible_description": "OpenAI APIのURL",
    "ai_config_openai_api_key_input_placeholder_text": "API キー",
    "ai_config_openai_api_key_input_accessible_description": "OpenAI APIのキー",
    "ai_config_openai_api_supported_models_label": "サポートされているモデル",
    "ai_config_ai_model_names_combo_placeholder_text": "モデルを選択",
    "ai_config_ai_model_names_combo_accessible_description": "選択可能なサポートされているモデル",

    "ai_config_base_label": "基本パラメータ",
    "ai_config_base_system_prompt_label": "システムプロンプト",
    "ai_config_base_system_prompt_edit_placeholder_text": "各リクエストに先立つ基本システムプロンプト",
    "ai_config_base_system_prompt_edit_accessible_description":
        "各リクエストに先立つ基本システムプロンプト。プレーンテキスト。",
    "ai_config_base_response_max_tokens_label": "応答の最大トークン数",
    "ai_config_base_response_max_tokens_input_accessible_description": "応答で受け取る最大トークン数",
}


from .enum_base import EnumBase

assert hasattr(EnumBase, 'default'), "Check default() method is implemented in the base class"


class AiModelNames(EnumBase):

    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        """
        Override _generate_next_value_ to allow for custom attributes on enum members (mostly for default).

        'name' is the name of the Enum's member, which will be 'A', 'B', 'C', etc.
        This method returns the name itself, a 'legacy' marker (False), and a 'default' marker (False)

        Legacy means the outdated model, e.g. legacy completions that is still supported.
        """
        return name, False, False

    GPT_3_5 = ("gpt-3.5-turbo", False, True)  # The third item in the tuple marks this as the default
    GPT_3_5_LEGACY = ("gpt-3.5-turbo-instruct", True)  # Second param is a legacy flag (e.g. legacy completions)

    def __init__(self, value, legacy=False, is_default=False):
        # If the value is a tuple, unpack it.
        if isinstance(value, tuple):
            self._value_, self.legacy, self.is_default = value
        else:
            self._value_ = value
            self.legacy = legacy
            self.is_default = is_default

    @classmethod
    def legacy_members(cls):
        return [member for member in cls if member.legacy]

    @classmethod
    def legacy(cls, member):
        return member in cls.legacy_members()



from setuptools import setup, find_packages
import pathlib

# Read the contents of README file
here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='notolog',
    version='0.9.1b4',
    description='Notolog - Python Markdown Editor',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/notolog/notolog-editor',
    author='Vadim Bakhrenkov',
    license='MIT',
    # py_modules=['main'],  # This refers to the standalone 'main.py' at the top-level directory
    entry_points={
        'console_scripts': [
            # 'notolog=main:main',  # main.py at the top-level directory
            'notolog=notolog.main:main',  # main.py within the main package
        ],
    },
    classifiers=[
        # More info https://pypi.org/classifiers/
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Text Editors',
        'Topic :: Text Editors :: Documentation',
        'Topic :: Text Editors :: Emacs',
        'Topic :: Text Editors :: Integrated Development Environments (IDE)',
        'Topic :: Text Editors :: Text Processing',
        'Topic :: Text Editors :: Word Processors',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: General',
        'Topic :: Text Processing :: Markup :: Markdown',
    ],
    packages=find_packages(),  # Find all packages in the directory
    include_package_data=True,  # Include data files specified in MANIFEST.in
    python_requires='>=3.9, <4',
    install_requires=[line.strip() for line in open("requirements.txt", "r")],
    project_urls={
        'Bug Reports': 'https://github.com/notolog/notolog-editor/issues',
        'Source': 'https://github.com/notolog/notolog-editor/',
    },
    keywords='notolog, python, markdown, editor, ai, text',
)



# Dutch lexemes common.py
lexemes = {
    "app_title": "Notolog Editor",
    "app_title_with_sub": "{app_title} - {sub_title}",

    "tree_filter_accessible_desc": "Bestandsfilterveld",

    "menu_action_rename": "Hernoemen",
    "menu_action_delete": "Verwijderen",
    "menu_action_delete_completely": "Volledig verwijderen",
    "menu_action_restore": "Herstellen",

    "dialog_file_rename_title": "Bestand hernoemen",
    "dialog_file_rename_field_label": "Voer nieuwe bestandsnaam in",
    "dialog_file_rename_button_ok": "Hernoemen",
    "dialog_file_rename_warning_exists": "Bestand met dezelfde naam bestaat al",

    "dialog_file_delete_title": "Bestand verwijderen",
    "dialog_file_delete_text": "Bestand \"{file_name}\" verwijderen?",
    "dialog_file_delete_completely_title": "Bestand volledig verwijderen",
    "dialog_file_delete_completely_text": "Bestand \"{file_name}\" volledig verwijderen?",
    "dialog_file_delete_error": "Kan bestand niet verwijderen, er is een fout opgetreden",
    "dialog_file_delete_error_not_found": "Bestand niet gevonden",

    "dialog_file_restore_title": "Bestand herstellen",
    "dialog_file_restore_text": "Bestand \"{file_name}\" herstellen?",
    "dialog_file_restore_error": "Kan bestand niet herstellen, er is een fout opgetreden",
    "dialog_file_restore_warning_exists": "Bestand met de naam {file_name} bestaat al",

    "dialog_message_box_title": "Bericht",
    "dialog_message_box_button_ok": "Sluiten",

    "action_new_file_first_line_template_text": "Nieuw document",
    "action_open_file_dialog_caption": "Bestand openen",
    "action_save_as_file_dialog_caption": "Bestand opslaan",

    "dialog_save_empty_file_title": "Leeg bestand opslaan",
    "dialog_save_empty_file_text": "Toestaan om het bestand met lege inhoud op te slaan?",

    "dialog_encrypt_file_title": "Bestand versleutelen",
    "dialog_encrypt_file_text": "Bestand \"{file_name}\" versleutelen?",
    "encrypt_file_warning_file_is_already_encrypted": "Het bestand is al versleuteld!",
    "dialog_encrypt_file_rewrite_existing_title": "Bestaand bestand overschrijven",
    "dialog_encrypt_file_rewrite_existing_text": "Bestaand bestand \"{file_path}\" overschrijven?",

    "dialog_decrypt_file_title": "Bestand ontsleutelen",
    "dialog_decrypt_file_text": "Bestand \"{file_name}\" ontsleutelen?",
    "decrypt_file_warning_file_is_not_encrypted": "Het bestand is niet versleuteld!",
    "dialog_decrypt_file_rewrite_existing_title": "Bestaand bestand herschrijven",
    "dialog_decrypt_file_rewrite_existing_text": "Bestaand bestand \"{file_path}\" herschrijven?",

    "dialog_encrypt_new_password_title": "Nieuw wachtwoord",
    "dialog_encrypt_new_password_label": "Wachtwoord:",
    "dialog_encrypt_new_password_input_placeholder_text": "Nieuw wachtwoord invoeren",
    "dialog_encrypt_new_password_hint_label": "Hint:",
    "dialog_encrypt_new_password_hint_label_description": "De hint is niet versleuteld en kan uit het bestand worden gelezen!"
                                                          "\nVermijd voor de hand liggende hints die gemakkelijk geraden kunnen worden, zoals"
                                                          "\ngeboortedata."
                                                          "\nProbeer een referentie te gebruiken die niet gemakkelijk met u in verband wordt gebracht.",
    "dialog_encrypt_new_password_hint_input_placeholder_text": "Hint invoeren (optioneel)",
    "dialog_encrypt_new_password_button_ok": "OK",
    "dialog_encrypt_new_password_button_cancel": "Annuleren",
    "dialog_encrypt_new_password_warning_empty_title": "Waarschuwing",
    "dialog_encrypt_new_password_warning_empty_text": "Het wachtwoordveld mag niet leeg zijn!",
    "dialog_encrypt_new_password_warning_too_long_title": "Waarschuwing",
    "dialog_encrypt_new_password_warning_too_long_text": "Het hintveld is te lang, maximaal {symbols} tekens!",

    "dialog_encrypt_password_title": "Wachtwoord invoeren",
    "dialog_encrypt_password_label": "Wachtwoord:",
    "dialog_encrypt_password_input_placeholder_text": "Wachtwoord invoeren",
    "dialog_encrypt_password_hint_label": "Hint:",
    "dialog_encrypt_password_button_ok": "OK",
    "dialog_encrypt_password_button_cancel": "Annuleren",

    "dialog_encrypt_password_reset_title": "Versleutelingswachtwoord resetten",
    "dialog_encrypt_password_reset_text": "Weet u zeker dat u het huidige versleutelingswachtwoord wilt resetten?",
    "dialog_encrypt_password_reset_button_cancel": "Annuleren",
    "dialog_encrypt_password_reset_button_yes": "Ja",

    "dialog_open_link_title": "Link",
    "dialog_open_link_text": "Link \"{url}\" openen in een browser?",

    "load_file_encryption_password_mismatch": "Versleutelingswachtwoord komt niet overeen!",
    "load_file_encryption_password_incorrect": "Onjuist versleutelingswachtwoord!",
    "load_file_none_content_error": "Bestand kan niet worden geladen.",

    "save_active_file_error_occurred": "Kan bestand niet opslaan, er is een fout opgetreden",

    "expandable_block_default_title": "Meer info...",

    "dialog_color_picker_color_copied_to_the_clipboard": "Geformatteerde tekst is naar het klembord gekopieerd",

    "popup_about_title": "Applicatie-info",
    "popup_about_app_name_description": "Python Markdown Editor",

    "popup_about_version": "Versie",
    "popup_about_license": "Licentie",
    "popup_about_website": "Website",
    "popup_about_repository": "GitHub",
    "popup_about_pypi": "PyPi",
    "popup_about_date": "Datum",

    "update_helper_new_version_is_available": "Een nieuwe versie '{latest_version}' van de app is beschikbaar",
    "update_helper_latest_version_installed": "De nieuwste versie van de app is geïnstalleerd",

    "network_connection_error_empty": "Kan geen responsinformatie verkrijgen",
    "network_connection_error_connection_or_dns":
        "Host niet gevonden. Er kan een probleem zijn met de internetverbinding of DNS.",
    "network_connection_error_connection_refused":
        "Verbinding geweigerd. De server kan uit zijn, of er kunnen netwerkproblemen zijn.",
    "network_connection_error_connection_timed_out": "Verbindingstijd verstreken. Er kunnen netwerkproblemen zijn.",
    "network_connection_error_connection_404_error":
        "404 fout bij verbinding. De gevraagde pagina of bron is niet gevonden.",
    "network_connection_error_generic_with_status_code": "Verzoek mislukt met statuscode: {status_code}",
}



from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDialog, QLabel, QLineEdit, QPushButton, QHBoxLayout, QVBoxLayout, QStyle, QMessageBox
from PySide6.QtGui import QFontMetrics, QFont

from . import Settings
from . import AppConfig
from . import Lexemes
from . import ThemeHelper

import logging


class EncNewPasswordDialog(QDialog):

    HINT_MAX_LENGTH = 32

    def __init__(self, parent=None):
        super().__init__(parent)

        self.parent = parent

        if self.parent and hasattr(self.parent, 'font'):
            # Apply font from the main window to the dialog
            self.setFont(self.parent.font())

        self.settings = Settings(parent=self)

        self.logger = logging.getLogger('enc_new_password_dialog')

        self.logging = AppConfig.get_logging()
        self.debug = AppConfig.get_debug()

        # Default language setup, change to settings value to modify it via UI
        self.lexemes = Lexemes(self.settings.app_language, default_scope='common')

        self.theme_helper = ThemeHelper()

        title = self.lexemes.get('dialog_encrypt_new_password_title')
        self.setWindowTitle(title)

        password_layout = QVBoxLayout()
        password_layout.addWidget(QLabel(self.lexemes.get('dialog_encrypt_new_password_label')))
        self.password_edit = QLineEdit()
        self.password_edit.setPlaceholderText(self.lexemes.get('dialog_encrypt_new_password_input_placeholder_text'))
        self.password_edit.setEchoMode(QLineEdit.EchoMode.Password)
        self.password_edit.setFocus()
        password_layout.addWidget(self.password_edit)

        hint_layout = QVBoxLayout()
        self.hint_edit = QLineEdit()
        self.hint_edit.setMaxLength(self.HINT_MAX_LENGTH)  # set maximum length of the input text
        hint_layout.addWidget(QLabel(self.lexemes.get('dialog_encrypt_new_password_hint_label')))
        hint_description = QLabel(self.lexemes.get('dialog_encrypt_new_password_hint_label_description'))
        hint_description.setFont(QFont(hint_description.font().family(), italic=True))
        hint_description.setStyleSheet("color: %s" % self.theme_helper.get_color('dialog_encrypt_new_password_hint',
                                                                                 css_format=True))
        self.set_font_size(hint_description, 85)
        hint_layout.addWidget(hint_description)
        self.hint_edit.setPlaceholderText(self.lexemes.get('dialog_encrypt_new_password_hint_input_placeholder_text'))
        hint_layout.addWidget(self.hint_edit)

        buttons_layout = QHBoxLayout()

        self.ok_button = QPushButton(self.lexemes.get('dialog_encrypt_new_password_button_ok'))
        self.ok_button.setIcon(self.style().standardIcon(QStyle.StandardPixmap.SP_DialogOkButton))
        self.ok_button.clicked.connect(self.validate_and_accept)  # or `self.accept`
        self.ok_button.setEnabled(False)

        # https://doc.qt.io/qt-6/qt.html#FocusPolicy-enum
        self.cancel_button = QPushButton(self.lexemes.get('dialog_encrypt_new_password_button_cancel'),
                                         focusPolicy=Qt.FocusPolicy.NoFocus)
        self.cancel_button.setIcon(self.style().standardIcon(QStyle.StandardPixmap.SP_DialogCancelButton))
        self.cancel_button.clicked.connect(self.reject)

        buttons_layout.addWidget(self.cancel_button)
        buttons_layout.addWidget(self.ok_button)

        layout = QVBoxLayout()
        layout.addLayout(password_layout)
        layout.addLayout(hint_layout)
        layout.addLayout(buttons_layout)

        self.setLayout(layout)

        self.adjust_minimum_width(title)

        self.password_edit.textChanged.connect(self.check_input_length)
        self.hint_edit.textChanged.connect(self.check_input_length)

    def check_input_length(self):
        password_length = len(self.password_edit.text())
        hint_length = len(self.hint_edit.text())

        # Enable OK button only if both fields have non-zero lengths
        self.ok_button.setEnabled(
            password_length > 0
            and (hint_length and 0 < hint_length <= self.HINT_MAX_LENGTH
                 or hint_length == 0)
        )

    def validate_and_accept(self):
        password = self.password_edit.text()
        hint = self.hint_edit.text()

        if not password:
            QMessageBox.warning(self,
                                self.lexemes.get('dialog_encrypt_new_password_warning_empty_title'),
                                self.lexemes.get('dialog_encrypt_new_password_warning_empty_text'))
            return

        if len(hint) > self.HINT_MAX_LENGTH:
            # This statement shouldn't be reachable anyway
            QMessageBox.warning(self,
                                self.lexemes.get('dialog_encrypt_new_password_warning_too_long_title'),
                                self.lexemes.get('dialog_encrypt_new_password_warning_too_long_text',
                                                 symbols=self.HINT_MAX_LENGTH))
            return

        self.accept()

    def adjust_minimum_width(self, title):
        # Calculate width required for the title text
        font_metrics = QFontMetrics(self.font())
        title_width = font_metrics.horizontalAdvance(title)

        # Set minimum width for the dialog
        self.setMinimumWidth(title_width + 150)  # Add some padding

    def set_font_size(self, widget, percent):
        default_font_size = self.font().pointSizeF()
        font = widget.font()
        font.setPointSizeF(default_font_size * (percent / 100))
        widget.setFont(font)

    """
    Example of class standalone usage:
    def dialog_show():
        dialog = EncNewPasswordDialog()
        if dialog.exec() == QDialog.DialogCode.Accepted:
            return dialog.password_edit.text(), dialog.hint_edit.text()
        return None, None

    if __name__ == "__main__":
        app = QApplication(sys.argv)
        password, hint = dialog_show()
        print("Password:", password)
        print("Hint:", hint)
        sys.exit(app.exec())
    """



# Portuguese lexemes settings_dialog.py
lexemes = {
    # Configurações
    "window_title": "Configurações",

    "button_close": "Fechar",

    "tab_general": "Geral",
    "tab_editor_config": "Editor",
    "tab_viewer_config": "Visualizador",
    "tab_ai_config": "Configuração de IA",

    "general_app_config_label": "Configuração do aplicativo",
    "general_app_language_label": "Idioma",
    "general_app_language_combo_placeholder_text": "Escolher um idioma",
    "general_app_language_combo_accessible_description": "Idioma da interface do aplicativo",
    "general_app_theme_label": "Tema",
    "general_app_theme_combo_placeholder_text": "Escolher um tema",
    "general_app_theme_combo_accessible_description": "Tema da interface do aplicativo",
    "general_app_main_menu_label": "Menu Principal",
    "general_app_main_menu_checkbox": "Mostrar menu principal",
    "general_app_main_menu_checkbox_accessible_description": "Exibir o menu dropdown principal do aplicativo",
    "general_app_font_size_label": "Tamanho da fonte: {size}pt",
    "general_app_font_size_slider_accessible_description": "Ajustar o tamanho da fonte global do aplicativo",

    "general_statusbar_label": "Barra de Status",
    "general_statusbar_show_global_cursor_position_checkbox": "Mostrar Posição Global do Cursor",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Exibir a posição global do cursor na barra de status",

    "editor_config_label": "Configuração do Editor",
    "editor_config_show_line_numbers_checkbox": "Mostrar Números de Linha",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Exibir números de linha no editor",

    "viewer_config_label": "Configuração do Visualizador",
    "viewer_config_process_emojis_checkbox": "Converter Emojis de Texto em Gráficos",
    "viewer_config_process_emojis_checkbox_accessible_description":
        "Converter emojis de texto em representações gráficas",
    "viewer_config_highlight_todos_checkbox": "Destacar TODOs",
    "viewer_config_highlight_todos_checkbox_accessible_description": "Destacar tags TODO no texto",
    "viewer_config_open_link_confirmation_checkbox": "Requerer Confirmação para Abrir Links",
    "viewer_config_open_link_confirmation_checkbox_accessible_description":
        "Solicitar confirmação antes de abrir links",
    "viewer_config_save_resources_checkbox": "Salvar automaticamente imagens externas no disco",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Salva automaticamente cópias de imagens externas no disco para acesso offline.",

    "ai_config_openai_api_label": "API da OpenAI",
    "ai_config_openai_api_url_input_placeholder_text": "URL da API",
    "ai_config_openai_api_url_input_accessible_description": "URL da API da OpenAI",
    "ai_config_openai_api_key_input_placeholder_text": "Chave da API",
    "ai_config_openai_api_key_input_accessible_description": "Chave da API da OpenAI",
    "ai_config_openai_api_supported_models_label": "Modelos suportados",
    "ai_config_ai_model_names_combo_placeholder_text": "Escolher modelo",
    "ai_config_ai_model_names_combo_accessible_description": "Modelos suportados para escolha",

    "ai_config_base_label": "Parâmetros Base",
    "ai_config_base_system_prompt_label": "Prompt do Sistema",
    "ai_config_base_system_prompt_edit_placeholder_text": "Prompt do sistema base que antecede cada solicitação",
    "ai_config_base_system_prompt_edit_accessible_description":
        "Prompt do sistema base que antecede cada solicitação. Texto simples.",
    "ai_config_base_response_max_tokens_label": "Máximo de Tokens de Resposta",
    "ai_config_base_response_max_tokens_input_accessible_description": "Número máximo de tokens a receber em resposta",
}


from PySide6.QtCore import Qt, QTimer
from PySide6.QtWidgets import QLabel
from PySide6.QtGui import QPixmap, QTransform, QPainter


class RotatingLabel(QLabel):
    def __init__(self, pixmap: QPixmap, parent=None):
        super().__init__(parent)

        self.rotation_angle = 0

        # Or just: self.pixmap = QPixmap("pixmap.png")
        self.pixmap = pixmap

        self.timer = QTimer(self)
        self.timer.timeout.connect(self.update_rotation)
        self.timer.start(50)  # Rotate every X milliseconds

    def setPixmap(self, pixmap):
        super().setPixmap(pixmap)
        self.pixmap = pixmap

    def paintEvent(self, event) -> None:
        """
        Works when self.show() triggered
        @param event: QPainter event
        @return: None
        """
        painter = QPainter(self)
        transform = QTransform().rotate(self.rotation_angle)
        rotated_pixmap = self.pixmap.transformed(transform, Qt.TransformationMode.SmoothTransformation)
        painter.drawPixmap(self.rect(), rotated_pixmap)

    def update_rotation(self):
        self.rotation_angle += 5  # Rotate by X degrees
        self.rotation_angle %= 360  # Ensure rotation angle stays within the range
        self.update()
